from typing import Union
from datalogue.errors import DtlError, _property_not_found
from datalogue.models.transformations.commons import Transformation


class MapFunction(Transformation):
    """
    Applies a given function to an ADG that should return and ADG

    It should be of the form:

    (adg: ADG, env: MAP) => {
        // Here the content of function goes
    }

    ALPHA Feature
    """

    type_str = "MapFunction"

    def __init__(self, f: str):
        """
        :param f: function to be applied to the adg
        """
        Transformation.__init__(self, MapFunction.type_str)
        self.f = f

    def __eq__(self, other: "MapFunction"):
        if isinstance(self, other.__class__):
            return self._as_payload() == other._as_payload()
        return False

    def __repr__(self):
        return f"MapFunction(f= {self.f!r})"

    def _as_payload(self) -> dict:
        base = self._base_payload()
        base["f"] = self.f
        return base

    @staticmethod
    def _from_payload(json: dict) -> Union[DtlError, "MapFunction"]:
        f = json.get("f")
        if f is None:
            return _property_not_found("f", json)

        return MapFunction(f)



import base64

from datalogue.clients._http import _HttpClient, Union, Optional, HttpMethod
from datalogue.clients.jobs import JobsClient
from datalogue.clients._datastore_collections import _DatastoreCollectionClient
from datalogue.clients.data_product import DataProductClient
from datalogue.clients.credentials import CredentialsClient
from datalogue.clients.datastore import DatastoreClient
from datalogue.clients._organization import OrganizationClient
from datalogue.clients._group import GroupClient
from datalogue.clients._tag import _TagClient
from datalogue.clients._user import UserClient
from datalogue.clients.ontology import OntologyClient
from datalogue.clients._classifier import _ClassifierClient
from datalogue.clients._regex import _RegexClient
from datalogue.clients.pipeline import PipelineClient
from datalogue.credentials import DtlCredentials, _get_ssl_verify_env
from datalogue.clients._authentication import _AuthenticationClient
from datalogue.errors import DtlError
from datalogue.version import __version__ as dtl_version

from uuid import UUID

from datalogue.models.organization import User


class LoginCredentials:
    def __init__(self, uri: str):
        self.uri = uri


class DtlTokenCredentials(LoginCredentials):
    def __init__(
        self,
        uri: str,
        user_id: Union[str, UUID],
        token: str,
        verify_certificate: bool = True,
    ):
        LoginCredentials.__init__(self, uri)
        self.user_id = user_id
        self.token = token
        self.verify_certificate = verify_certificate


class Dtl:
    """
    Root class to be built to interact with all the services

    :param credentials: contains the information to connect
    """

    def __init__(self, credentials: LoginCredentials):
        self.uri = credentials.uri

        if isinstance(credentials, DtlCredentials):
            self.username = credentials.username
            self.http_client = _HttpClient(
                credentials.uri,
                credentials.authentication_name,
                credentials.verify_certificate,
            )

            login_res = self.http_client.login(
                credentials.username, credentials.password
            )
            if isinstance(login_res, DtlError):
                raise login_res

        if isinstance(credentials, DtlTokenCredentials):
            user_id_with_token = f"{credentials.user_id}:{credentials.token}"
            encoded_bytes = base64.b64encode(user_id_with_token.encode("utf-8"))
            encoded_token = str(encoded_bytes, "utf-8")

            self.http_client = _HttpClient(
                credentials.uri,
                verify_certificate=credentials.verify_certificate,
                encoded_auth_header=f"TOKEN {encoded_token}",
            )
            current_user = UserClient(self.http_client).get_current_user()

            if isinstance(current_user, DtlError):
                raise current_user

            self.username = current_user.email

        self.group = GroupClient(self.http_client)
        """Client to interact with the groups"""
        self.user = UserClient(self.http_client)
        """Client to interact with the users"""
        self.organization = OrganizationClient(self.http_client)
        """Client to interact with the organization part of the stack"""
        self.jobs = JobsClient(self.http_client)
        """Client to interact with the jobs"""
        self.datastore_collection = _DatastoreCollectionClient(self.http_client)
        """Client to interact with the datastore collections"""
        self.datastore = DatastoreClient(self.http_client)
        """Client to interact with the datastores"""
        self.credentials = CredentialsClient(self.http_client)
        """Client to interact with credentials"""
        self.ontology = OntologyClient(self.http_client)
        """Client to interact with ontologies"""
        self.classifier = _ClassifierClient(self.http_client, self.ontology)
        """Client to interact with classifiers"""
        self.regex = _RegexClient(self.http_client)
        """Client to interact with regexes"""
        self.tag = _TagClient(self.http_client)
        """Client to interact with tags"""
        self.pipeline = PipelineClient(self.http_client)
        """Client to interact with the stream collections"""
        self.authentication = _AuthenticationClient(self.http_client)
        """Client to interact with the Authentication Schemes"""
        self.data_product = DataProductClient(self.http_client)
        """Client to interact with the data products"""

    def __repr__(self):
        return f"Logged in {self.uri!r} with {self.username!r} account."

    @staticmethod
    def signup(
        uri="",
        first_name="",
        last_name="",
        email="",
        password="",
        accept_terms=True,
        verify_certificate: Optional[bool] = None,
    ) -> Union[DtlError, User]:
        """
        Perform signup of user
        :param uri: The target URI where the user data will be associated in
        :param accept_terms: Whether the user accept the following terms : https://www.datalogue.io/pages/terms-of-service
        :return: User object if successful, else return DtlError
        """

        if verify_certificate is None:
            verify_certificate = _get_ssl_verify_env()

        http_client = _HttpClient(uri, verify_certificate=verify_certificate)
        http_client.get_csrf()
        registered_user = http_client.signup(
            first_name, last_name, email, password, accept_terms
        )
        return registered_user

    def version(self) -> str:
        """
        Get version information of SDK
        :return: Build number of the SDK
        """
        return dtl_version



from abc import ABC, abstractmethod
from typing import Dict, List, Set, Optional, Union, NamedTuple, Literal
from uuid import UUID
from datetime import datetime
from dateutil.tz import UTC
from datalogue.errors import _enum_parse_error, DtlError
from datalogue.models.kinesis.shards import ShardAttributes
from datalogue.models.schema import AbstractDataSchema, SchemaNode, Classification
from datalogue.models.pipeline_behavior import (
    PipelineBehavior,
    _pipeline_behavior_from_payload,
    SourceBehavior,
    _source_behavior_from_payload,
)
from datalogue.dtl_utils import (
    _parse_list,
    _parse_string_list,
    SerializableStringEnum,
    _parse_uuid,
)
from datalogue.dtl_utils import map_option, _parse_datetime


class NodeClassification:
    """
    The mapping of a class to a field, as applied using the add, remove, and override commands of the dtl.datastore client.

    :param schema_node_id: the id of a schema field
    :param ontology_node_id: the class of a schema field
    :param score: the Confidence score for the classification, from 0.0 - 1.0 (defaults to 1.0)
    """

    def __init__(
        self,
        schema_node_id: Union[str, UUID],
        ontology_node_id: Union[str, UUID],
        score: float = 1.0,
    ):
        self.schema_node_id = schema_node_id
        self.ontology_node_id = ontology_node_id
        self.score = score

    def __repr__(self):
        return (
            f"{self.__class__.__name__}("
            f"schema_node_id= {self.schema_node_id!r}, "
            f"ontology_node_id= {self.ontology_node_id!r},"
            f"score= {self.score!r}, "
        )

    def _as_payload(self) -> Union[DtlError, dict]:
        """
        Dictionary representation of the object with camelCase keys
        :return:
        """
        return {
            "schemaNodeId": str(self.schema_node_id),
            "classId": str(self.ontology_node_id),
            "score": str(self.score),
        }


class CredentialType(SerializableStringEnum):
    S3 = "S3"
    AmazonVendorCentral = "AmazonVendorCentral"
    GCS = "GCS"
    Azure = "Azure"
    JDBC = "Jdbc"
    Mongo = "Mongo"
    Http = "Http"
    Socrata = "Socrata"
    FileSystem = "FileSystem"
    Void = "Void"
    Kinesis = "Kinesis"
    Hadoop = "Hadoop"
    Kafka = "Kafka"
    Minio = "Minio"
    HadoopAzure = "HadoopAzure"
    Webhook = "Webhook"
    SFTP = "SFTP"
    Kerberos = "Kerberos"
    JaasString = "Kerberos"
    Neo4j = "Neo4j"
    ElasticSearch = "ElasticSearch"
    Box = "Box"

    @staticmethod
    def parse_error(s: str) -> DtlError:
        return DtlError(_enum_parse_error("store type", s))

    @staticmethod
    def credential_type_from_str(string: str) -> Union[DtlError, "CredentialType"]:
        return SerializableStringEnum.from_str(CredentialType)(string)


class EncryptionType(SerializableStringEnum):
    AES256 = "AES256"
    KeyId = "KeyId"
    CustomerKey = "CustomerKey"

    @staticmethod
    def parse_error(s: str) -> DtlError:
        return DtlError(_enum_parse_error("encryption type", s))

    @staticmethod
    def encryption_type_from_str(string: str) -> Union[DtlError, "EncryptionType"]:
        return SerializableStringEnum.from_str(EncryptionType)(string)


class FileFormat(SerializableStringEnum):
    Json = "Json"
    Csv = "Csv"
    Text = "Text"
    Xml = "Xml"
    Html = "Html"
    Excel = "Excel"
    Avro = "Avro"
    Parquet = "Parquet"

    @staticmethod
    def parse_error(s: str) -> DtlError:
        return DtlError(_enum_parse_error("file format", s))

    @staticmethod
    def file_format_from_str(string: str) -> Union[DtlError, "FileFormat"]:
        return SerializableStringEnum.from_str(FileFormat)(string)


class HttpMethodType(SerializableStringEnum):
    PUT = "PUT"
    PATCH = "PATCH"
    POST = "POST"
    GET = "GET"

    @staticmethod
    def parse_error(s: str) -> DtlError:
        return DtlError(_enum_parse_error("file format", s))

    @staticmethod
    def method_type_from_str(string: str) -> Union[DtlError, "HttpMethodType"]:
        return SerializableStringEnum.from_str(HttpMethodType)(string)


class ResourceType(SerializableStringEnum):
    EventDirectory = "EventDirectory"
    SnapshotDirectory = "SnapshotDirectory"
    Unknown = "Unknown"
    File = "File"

    @staticmethod
    def parse_error(s: str) -> DtlError:
        return DtlError(_enum_parse_error("resource type", s))

    @staticmethod
    def resource_type_from_str(string: str) -> Union[DtlError, "ResourceType"]:
        return SerializableStringEnum.from_str(ResourceType)(string)


class Decompression(SerializableStringEnum):
    Gzip = "Gzip"
    Infer = "Infer"
    NonGzip = "None"

    @staticmethod
    def parse_error(s: str) -> DtlError:
        return DtlError(_enum_parse_error("decompression type", s))

    @staticmethod
    def decompression_from_str(string: str) -> Union[DtlError, "Decompression"]:
        return SerializableStringEnum.from_str(Decompression)(string)


class SchemaFormat(SerializableStringEnum):
    Avro = "Avro"
    Json = "Json"
    Protobuf = "Protobuf"

    @staticmethod
    def parse_error(s: str) -> DtlError:
        return DtlError(_enum_parse_error("schema format", s))

    @staticmethod
    def schema_format_from_str(string: str) -> Union[DtlError, "SchemaFormat"]:
        return SerializableStringEnum.from_str(SchemaFormat)(string)


class ChangeDataSource(SerializableStringEnum):
    InformaticaFlat = "InformaticaFlat"
    InformaticaNested = "InformaticaNested"
    InformaticaGeneric = "InformaticaGeneric"
    InformaticaWrapped = "InformaticaWrapped"

    @staticmethod
    def parse_error(s: str) -> DtlError:
        return DtlError(_enum_parse_error("Change Data Source", s))

    @staticmethod
    def change_data_source_from_str(string: str) -> Union[DtlError, "ChangeDataSource"]:
        return SerializableStringEnum.from_str(ChangeDataSource)(string)


class DatastoreDef(ABC):
    type_field = "type"
    id_field = "id"

    def __init__(self, credential_type: CredentialType, datastore_id: Optional[UUID] = None):
        self.type = credential_type
        self.datastore_id = datastore_id
        super().__init__()

    def __eq__(self, other: "DatastoreDef"):
        if isinstance(self, other.__class__):
            return self._as_payload() == other._as_payload()
        return False

    def _base_payload(self) -> dict:
        base = [(DatastoreDef.type_field, self.type.value)]
        if self.datastore_id is not None:
            base.append((DatastoreDef.id_field, str(self.datastore_id)))
        return dict(base)

    @abstractmethod
    def _as_payload(self) -> dict:
        """
        Dictionary representation of the object
        :return:
        """
        pass


class Cell:
    """
    Describes a cell in html table
    """

    def __init__(
        self,
        label: str,
        span: int,
        index: int,
        total: int,
        tag: str,
        tagOverrides: [] = None,
    ):
        """
        Builds a cell

        :param label: text to display
        :param span: the html column span attribute
        :param index: index of this cell in the row
        :param total: total number of cells in the row
        """
        self.label = label
        self.span = span
        self.index = index
        self.total = total
        self.tag = tag
        self.tagOverrides = tagOverrides

    def __repr__(self):
        return f"{self.__class__.__name__}(label= {self.label!r}, span= {self.span!r}, \
                index= {self.index!r}, total= {self.total!r}, tag= {self.tag!r}, tagOverrides= {self.tagOverrides!r})"


def _cell_from_payload(json: dict) -> Union[DtlError, Cell]:
    label = json.get("label")
    if label is None:
        return DtlError("Cell has to have a 'label' key")

    span = json.get("span")
    if span is None:
        return DtlError("Cell has to have a 'span' key")

    index = json.get("index")
    if index is None:
        return DtlError("Cell has to have an 'index' key")

    total = json.get("total")
    if total is None:
        return DtlError("Cell has to have a 'total' key")

    return Cell(label, span, index, total, json.get("tag"), json.get("tagOverrides"))


class Datastore:
    """
    Represents a pointer to a datastore
    """

    def __init__(
        self,
        name: str,
        definition: DatastoreDef,
        credential_id: Optional[UUID] = None,
        alias: Optional[str] = None,
        datastore_id: Optional[UUID] = None,
        schema_paths: Optional[List[List[str]]] = None,
        schema_classes: Optional[List[List[str]]] = None,
        schema_labels: Optional[Set[str]] = None,
        schema_nodes: Optional[List[SchemaNode]] = None,
        ads: Optional[AbstractDataSchema] = None,
        tags: List[str] = [],
    ):
        """
        Builds a pointer to a datastore

        :param name: name of the pointer
        :param alias: unique alias accross the organization for the pointer
        :param definition: definition on how to access the data
        :param credential_id: reference to the Credential to access the data
        :param datastore_id: defined if the object was persisted, otherwise None
        :param schema_paths: Used to replace paths options on the cli
        :param schema_classes: Used to replace classes options on the cli
        :param schema_labels: Set of labels that exists in schema
        :param tags: Set of labels that exists in schema
        """
        self.id = datastore_id
        self.name = name
        self.alias = alias
        self.definition = definition
        self.definition.datastore_id = datastore_id
        self.credential_id = credential_id
        self.schema_paths = schema_paths
        self.schema_classes = schema_classes
        self.schema_labels = schema_labels
        self.schema_nodes = schema_nodes
        self.ads = ads
        self.tags = tags

    def __eq__(self, other: "Datastore"):
        if isinstance(self, other.__class__):
            return self._as_payload() == other._as_payload()
        return False

    def __repr__(self):
        return (
            f"{self.__class__.__name__}(id= {self.id}, name= {self.name!r}, alias= {self.alias!r}, "
            f"credential_id= {self.credential_id}, definition= {self.definition!r}, "
            f"schema_paths= {self.schema_paths!r}, schema_labels= {self.schema_labels!r}, "
            f"schema_nodes= {self.schema_nodes!r}, tags= {self.tags})"
        )

    def _as_payload(self) -> dict:
        """
        Dictionary representation of the object
        :return:
        """
        base = {
            "name": self.name,
            "gate": self.definition._as_payload(),
        }
        if self.id is not None:
            base["id"] = str(self.id)

        if self.alias is not None:
            base["alias"] = self.alias

        if self.credential_id is not None:
            base["credentialsId"] = str(self.credential_id)

        return base


###############################################################################
#                              Cloud Storage
###############################################################################


class S3DatastoreDef(DatastoreDef):
    type_str = CredentialType.S3

    def __init__(
        self,
        bucket: str,
        key: str,
        file_format: FileFormat,
        params: Optional[Dict[str, str]] = None,
        datastore_id: Optional[UUID] = None,
    ):
        DatastoreDef.__init__(self, S3DatastoreDef.type_str, datastore_id)
        self.bucket = bucket
        self.key = key
        self.file_format = file_format
        self.params = params

    def __repr__(self):
        return (
            f"{self.__class__.__name__}(bucket= {self.bucket!r}, key= {self.key!r}, "
            f"file_format= {self.file_format!r}, params= {self.params})"
        )

    def _as_payload(self) -> dict:
        """
        Dictionary representation of the object with camelCase keys
        :return:
        """
        base = self._base_payload()
        base["bucket"] = self.bucket
        base["key"] = self.key
        base["format"] = self.file_format.value
        if self.params is not None:
            base["params"] = self.params
        return base


def _s3_datastore_def_from_payload(d: dict) -> Union[DtlError, S3DatastoreDef]:
    type_field = d.get(DatastoreDef.type_field)
    if not isinstance(type_field, str):
        return DtlError("string %s is missing from the json" % (DatastoreDef.type_field))

    if type_field != S3DatastoreDef.type_str.value:
        return DtlError("The object %s is not an S3 definition" % (str(d)))

    bucket = d.get("bucket")
    if bucket is None:
        return DtlError("'bucket' needs to be defined in an S3 definition")

    key = d.get("key")
    if key is None:
        return DtlError("'key' needs to be defined in an S3 definition")

    file_format = d.get("format")
    if file_format is None:
        return DtlError("'file_format' needs to be defined in an S3 definition")
    else:
        file_format = FileFormat.file_format_from_str(file_format)
        if isinstance(file_format, DtlError):
            return file_format

    params = d.get("params")
    datastore_id = d.get(DatastoreDef.id_field)

    return S3DatastoreDef(bucket, key, file_format, params, datastore_id)


class AVCOrderItemsDatastore(DatastoreDef):
    type_str = CredentialType.AmazonVendorCentral

    def __init__(
        self,
        marketplace_id: str,
        created_after: datetime,
        created_before: Optional[datetime],
        datastore_id: Optional[UUID] = None,
    ):
        DatastoreDef.__init__(self, AVCOrderItemsDatastore.type_str, datastore_id)
        self.marketplace_id = marketplace_id
        self.created_after = created_after
        self.created_before = created_before

    def __repr__(self):
        created_after_str = self.created_after.strftime("%Y-%m-%dT%H:%M:%SZ")
        created_before_str = None
        if self.created_before is not None:
            created_before_str = self.created_before.strftime("%Y-%m-%dT%H:%M:%SZ")
        return (
            f"AVCOrderItemsDatastore(marketplace_id= {self.marketplace_id!r}, created_after= {created_after_str!r}, "
            f"created_before= {created_before_str!r})"
        )

    def _as_payload(self) -> dict:
        """
        Dictionary representation of the object  with camelCase keys
        :return:
        """
        base = self._base_payload()
        base["marketplaceId"] = self.marketplace_id
        base["createdAfter"] = self.created_after.strftime("%Y-%m-%dT%H:%M:%SZ")
        if self.created_before is not None:
            base["createdBefore"] = self.created_before.strftime("%Y-%m-%dT%H:%M:%SZ")
        return base


def _avc_datastore_def_from_payload(d: dict) -> Union[DtlError, AVCOrderItemsDatastore]:
    type_field = d.get(DatastoreDef.type_field)
    if not isinstance(type_field, str):
        return DtlError("string %s is missing from the json" % (DatastoreDef.type_field))

    if type_field != AVCOrderItemsDatastore.type_str.value:
        return DtlError("The object %s is not a AVCOrderItemsDatastore definition" % (str(d)))

    marketplace_id = d.get("marketplaceId")
    if marketplace_id is None:
        return DtlError("'marketplaceId' needs to be defined in a AVCOrderItemsDatastore definition")

    created_after = map_option(d.get("createdAfter"), _parse_datetime)
    if created_after is None:
        return DtlError("'createdAfter' needs to be defined in a AVCOrderItemsDatastore definition")

    created_before = map_option(d.get("createdBefore"), _parse_datetime)
    store_id = d.get(DatastoreDef.id_field)

    return AVCOrderItemsDatastore(marketplace_id, created_after, created_before, store_id)


class GCSDatastoreDef(DatastoreDef):
    type_str = CredentialType.GCS

    def __init__(
        self,
        bucket: str,
        file_name: str,
        file_format: FileFormat,
        params: Optional[Dict[str, str]] = None,
        datastore_id: Optional[UUID] = None,
    ):
        DatastoreDef.__init__(self, GCSDatastoreDef.type_str, datastore_id)
        self.bucket = bucket
        self.file_name = file_name
        self.file_format = file_format
        self.params = params

    def __repr__(self):
        return (
            f"{self.__class__.__name__}(bucket= {self.bucket!r}, file_name= {self.file_name!r}, "
            f"file_format= {self.file_format!r}, params= {self.params})"
        )

    def _as_payload(self) -> dict:
        """
        Dictionary representation of the object  with camelCase keys
        :return:
        """
        base = self._base_payload()
        base["bucket"] = self.bucket
        base["fileName"] = self.file_name
        base["format"] = self.file_format.value
        if self.params is not None:
            base["params"] = self.params
        return base


def _gcs_datastore_def_from_payload(d: dict) -> Union[DtlError, GCSDatastoreDef]:
    type_field = d.get(DatastoreDef.type_field)
    if not isinstance(type_field, str):
        return DtlError("string %s is missing from the json" % (DatastoreDef.type_field))

    if type_field != GCSDatastoreDef.type_str.value:
        return DtlError("The object %s is not a GCS definition" % (str(d)))

    bucket = d.get("bucket")
    if bucket is None:
        return DtlError("'bucket' needs to be defined in a GCS definition")

    file_name = d.get("fileName")
    if file_name is None:
        return DtlError("'fileName' needs to be defined in a GCS definition")

    file_format = d.get("format")
    if file_format is None:
        return DtlError("'file_format' needs to be defined in a GCS definition")
    else:
        file_format = FileFormat.file_format_from_str(file_format)
        if isinstance(file_format, DtlError):
            return file_format

    params = d.get("params")
    store_id = d.get(DatastoreDef.id_field)

    return GCSDatastoreDef(bucket, file_name, file_format, params, store_id)


class AzureDatastoreDef(DatastoreDef):
    type_str = CredentialType.Azure

    def __init__(
        self,
        container: str,
        file_name: str,
        file_format: FileFormat,
        params: Optional[Dict[str, str]] = None,
        datastore_id: Optional[UUID] = None,
    ):
        DatastoreDef.__init__(self, AzureDatastoreDef.type_str, datastore_id)
        self.container = container
        self.file_name = file_name
        self.file_format = file_format
        self.params = params

    def __repr__(self):
        return (
            f"{self.__class__.__name__}(container= {self.container!r}, file_name= {self.file_name!r}, "
            f"file_format= {self.file_format!r}, params= {self.params})"
        )

    def _as_payload(self) -> dict:
        """
        Dictionary representation of the object with camelCase keys
        :return:
        """
        base = self._base_payload()
        base["container"] = self.container
        base["fileName"] = self.file_name
        base["format"] = self.file_format.value
        if self.params is not None:
            base["params"] = self.params
        return base


def _azure_datastore_def_from_payload(d: dict) -> Union[DtlError, AzureDatastoreDef]:
    type_field = d.get(DatastoreDef.type_field)
    if not isinstance(type_field, str):
        return DtlError("string %s is missing from the json" % DatastoreDef.type_field)

    if type_field != AzureDatastoreDef.type_str.value:
        return DtlError("The object %s is not an Azure Datastore definition" % (str(d)))

    container = d.get("container")
    if container is None:
        return DtlError("'container' needs to be defined in an Azure Datastore definition")

    file_name = d.get("fileName")
    if file_name is None:
        return DtlError("'fileName' needs to be defined in an Azure Datastore definition")

    file_format = d.get("format")
    if file_format is None:
        return DtlError("'file_format' needs to be defined in an Azure Datastore definition")
    else:
        file_format = FileFormat.file_format_from_str(file_format)
        if isinstance(file_format, DtlError):
            return file_format

    params = d.get("params")
    datastore_id = d.get(DatastoreDef.id_field)

    return AzureDatastoreDef(container, file_name, file_format, params, datastore_id)


class DatalogueDatastoreDef(DatastoreDef):
    type_str = CredentialType.Minio

    def __init__(
        self,
        bucket: str,
        key: str,
        file_format: FileFormat,
        params: Optional[Dict[str, str]] = None,
        datastore_id: Optional[UUID] = None,
    ):
        DatastoreDef.__init__(self, DatalogueDatastoreDef.type_str, datastore_id)
        self.bucket = bucket
        self.key = key
        self.file_format = file_format
        self.params = params

    def __repr__(self):
        return (
            f"{self.__class__.__name__}(bucket= {self.bucket!r}, key= {self.key!r}, "
            f"file_format= {self.file_format!r}, params= {self.params})"
        )

    def _as_payload(self) -> dict:
        """
        Dictionary representation of the object with camelCase keys
        :return:
        """
        base = self._base_payload()
        base["bucket"] = self.bucket
        base["key"] = self.key
        base["format"] = self.file_format.value
        if self.params is not None:
            base["params"] = self.params
        return base

    @staticmethod
    def _from_payload(d: dict) -> Union[DtlError, "DatalogueDatastoreDef"]:
        type_field = d.get(DatastoreDef.type_field)
        if not isinstance(type_field, str):
            return DtlError("string %s is missing from the json" % (DatastoreDef.type_field))

        if type_field != DatalogueDatastoreDef.type_str.value:
            return DtlError("The object %s is not an Minio definition" % (str(d)))

        bucket = d.get("bucket")
        if bucket is None:
            return DtlError("'bucket' needs to be defined in an Minio definition")

        key = d.get("key")
        if key is None:
            return DtlError("'key' needs to be defined in an Minio definition")

        file_format = d.get("format")
        if file_format is None:
            return DtlError("'file_format' needs to be defined in an Minio definition")
        else:
            file_format = FileFormat.file_format_from_str(file_format)
            if isinstance(file_format, DtlError):
                return file_format

        params = d.get("params")
        datastore_id = d.get(DatastoreDef.id_field)

        return DatalogueDatastoreDef(bucket, key, file_format, params, datastore_id)


class HadoopDatastoreDef(DatastoreDef):
    type_str = CredentialType.Hadoop

    def __init__(
        self,
        location: str,
        file_format: FileFormat,
        params: Optional[Dict[str, str]] = None,
        datastore_id: Optional[UUID] = None,
    ):
        DatastoreDef.__init__(self, HadoopDatastoreDef.type_str, datastore_id)
        self.location = location
        self.file_format = file_format
        self.params = params

    def __repr__(self):
        return (
            f"{self.__class__.__name__}(location= {self.location!r}, "
            f"file_format= {self.file_format!r}, params= {self.params})"
        )

    def _as_payload(self) -> dict:
        """
        Dictionary representation of the object with camelCase keys
        :return:
        """
        base = self._base_payload()
        base["location"] = self.location
        base["format"] = self.file_format.value
        if self.params is not None:
            base["params"] = self.params
        return base

    @staticmethod
    def _from_payload(d: dict) -> Union[DtlError, "HadoopDatastoreDef"]:
        type_field = d.get(DatastoreDef.type_field)
        if not isinstance(type_field, str):
            return DtlError("string %s is missing from the json" % (DatastoreDef.type_field))

        if type_field != HadoopDatastoreDef.type_str.value:
            return DtlError("The object %s is not a Hadoop definition" % (str(d)))

        location = d.get("location")
        if location is None:
            return DtlError("'location' needs to be defined")

        file_format = d.get("format")
        if file_format is None:
            return DtlError("'format' needs to be defined")
        else:
            file_format = FileFormat.file_format_from_str(file_format)
            if isinstance(file_format, DtlError):
                return file_format
            if file_format != FileFormat.Parquet and file_format != FileFormat.Avro:
                return DtlError("'format' needs to be Avro or Parquet")

        params = d.get("params")
        resource_id = d.get(DatastoreDef.id_field)

        return HadoopDatastoreDef(location, file_format, params, resource_id)


class SchemaRegistry(NamedTuple):
    format: SchemaFormat
    schema_for_write: Optional[str] = None

    @staticmethod
    def _as_payload(self) -> dict:
        """
        Dictionary representation of the object with camelCase keys
        :return:
        """
        base = {"format": self.format.value}
        if self.schema_for_write is not None:
            base["schemaForWrite"] = self.schema_for_write
        return base

    @staticmethod
    def _from_payload(d: dict) -> Union[DtlError, "SchemaRegistry"]:
        schema_for_write = d.get("schemaForWrite")

        schema_format = d.get("format")

        if schema_format is None:
            return DtlError("'format' needs to be defined")
        else:
            schema_format = SchemaFormat.schema_format_from_str(schema_format)
            if isinstance(schema_format, DtlError):
                return schema_format

        return SchemaRegistry(schema_format, schema_for_write)


###############################################################################
#
#                              Database Connectors
###############################################################################


class JdbcDatastoreDef(DatastoreDef):
    type_str = CredentialType.JDBC

    def __init__(
        self,
        root_table: Optional[str] = None,
        datastore_id: Optional[UUID] = None,
        view_query: Optional[str] = None,
        post_processing_query: Optional[str] = None,
        params: Optional[Dict[str, str]] = None,
        schema: Optional[str] = None,
        pipeline_behavior: Optional[PipelineBehavior] = None,
        read_behavior: Optional[SourceBehavior] = None,
    ):
        """
        Defines a JDBC based data source

        :param root_table: the table to capture with this datastore; required for post_processing_query
        :param view_query: can be used to query the database, using the result as the data for that datastore.
            Please use a syntax appropriate for your dB--e.g. SQL for Postgres, CQL for Cassandra, etc.
        :param post_processing_query: can be used to run a query against the root table after a pipeline is run
            targeting that table
        :param params: Passing extra params to the datastore, currently used for Databricks Spark JDBC. Please consult this document on how to use this field : https://docs.google.com/document/d/15odo_ykW_LLiG1RR4g073jZ_QgHiq0Gbnti9TVZS2J4/edit#heading=h.w6jqud4z6wvf
        :raise ValueError: raises value error when both root_table and view_query are not specified. Also when both
            root_table and view_query are specified.
        """
        DatastoreDef.__init__(self, JdbcDatastoreDef.type_str, datastore_id)
        if root_table is None and view_query is None:
            raise ValueError("Either root_table or view_query must be specified for a JDBC datasource")
        if root_table is not None and view_query is not None:
            raise ValueError("Both root_table and view_query cannot be specified together for a JDBC datastore")
        if post_processing_query is not None and root_table is None:
            raise ValueError("root_table has to be defined in order to use post_processing_query")
        self.root_table = root_table
        self.view_query = view_query
        self.post_processing_query = post_processing_query
        self.params = params
        self.schema = schema
        self.pipeline_behavior = pipeline_behavior
        self.read_behavior = read_behavior

    def __repr__(self):
        if self.root_table is None:
            return (
                f"JdbcDatastoreDef(root_table= {self.root_table!r}, "
                f"post_processing_query= {self.post_processing_query!r}, "
                f"pipeline_behavior= {self.pipeline_behavior!r}, "
                f"read_behavior= {self.read_behavior!r})"
            )
        else:
            return (
                f"JdbcDatastoreDef(view_query= {self.view_query!r}, "
                f"post_processing_query= {self.post_processing_query!r}, "
                f"pipeline_behavior= {self.pipeline_behavior!r}, "
                f"read_behavior= {self.read_behavior!r})"
            )

    def _as_payload(self) -> dict:
        """Dictionary representation of the object with camelCase keys"""
        base = self._base_payload()
        if self.schema is not None:
            base["schema"] = self.schema
        if self.root_table is not None:
            base["rootTable"] = self.root_table
        if self.view_query is not None:
            base["viewQuery"] = self.view_query
        if self.post_processing_query is not None:
            base["postProcessingQuery"] = self.post_processing_query
        if self.pipeline_behavior is not None:
            base["pipelineBehavior"] = self.pipeline_behavior._as_payload()
        if self.read_behavior is not None:
            base["readBehavior"] = self.read_behavior._as_payload()
        base["params"] = self.params
        return base


def _jdbc_datastore_def_from_payload(d: dict) -> Union[DtlError, JdbcDatastoreDef]:
    type_field = d.get(DatastoreDef.type_field)
    if not isinstance(type_field, str):
        return DtlError("string %s is missing from the json" % DatastoreDef.type_field)

    if type_field != JdbcDatastoreDef.type_str.value:
        return DtlError("The object %s is not a Jdbc definition" % str(d))

    datastore_id = d.get(DatastoreDef.id_field)

    root_table = d.get("rootTable")
    view_query = d.get("viewQuery")
    if root_table is None and view_query is None:
        return DtlError("'rootTable' or 'viewQuery' needs to be defined in a Jdbc definition")
    if root_table is not None and view_query is not None:
        return DtlError("Either 'rootTable' or 'viewQuery' has to be defined not both in a Jdbc definition")

    post_query = d.get("postProcessingQuery")
    params = d.get("params")
    schema = d.get("schema")

    pipeline_behavior = d.get("pipelineBehavior")
    if pipeline_behavior is not None:
        pipeline_behavior = _pipeline_behavior_from_payload(pipeline_behavior)
    read_behavior = d.get("readBehavior")
    if read_behavior is not None:
        read_behavior = _source_behavior_from_payload(read_behavior)
    return JdbcDatastoreDef(
        root_table,
        datastore_id,
        view_query,
        post_query,
        params,
        schema,
        pipeline_behavior,
        read_behavior,
    )


class MongoDatastoreDef(DatastoreDef):
    type_str = CredentialType.Mongo

    def __init__(self, collection: str, datastore_id: Optional[UUID] = None):
        DatastoreDef.__init__(self, MongoDatastoreDef.type_str, datastore_id)
        self.collection = collection

    def _as_payload(self) -> dict:
        """Dictionary representation of the object with camelCase keys"""
        base = self._base_payload()
        base["collection"] = self.collection
        return base


def _mongo_datastore_def_from_payload(d: dict) -> Union[DtlError, MongoDatastoreDef]:
    type_field = d.get(DatastoreDef.type_field)
    if not isinstance(type_field, str):
        return DtlError("string %s is missing from the json" % DatastoreDef.type_field)

    if type_field != MongoDatastoreDef.type_str.value:
        return DtlError("The object %s is not a Mongo definition" % str(d))

    collection = d.get("collection")
    if collection is None:
        return DtlError("'collection' needs to be defined in a Mongo definition")

    datastore_id = d.get(DatastoreDef.id_field)

    return MongoDatastoreDef(collection, datastore_id)


class KinesisDatastoreDef(DatastoreDef):
    """
    Creates a datastore that references a Kinesis dataset
    """

    type_str = CredentialType.Kinesis

    def __init__(
        self,
        stream_name: str,
        file_format: FileFormat,
        shard_attributes: ShardAttributes = ShardAttributes(),
        params: Optional[Dict[str, str]] = None,
        datastore_id: Optional[UUID] = None,
    ):
        DatastoreDef.__init__(self, KinesisDatastoreDef.type_str, datastore_id)
        self.stream_name = stream_name
        self.file_format = file_format
        self.shard_attributes = shard_attributes
        self.params = params

    def __repr__(self):
        return (
            f"{self.__class__.__name__}(stream_name= {self.stream_name!r}, file_format= {self.file_format!r}, "
            f"shard_attributes= {self.shard_attributes!r}, params= {self.params})"
        )

    def _as_payload(self) -> dict:
        """
        Dictionary representation of the object with camelCase keys
        :return:
        """
        base = self._base_payload()
        base["streamName"] = self.stream_name
        base["format"] = self.file_format.value
        base["shardAttributes"] = self.shard_attributes._as_payload()
        if self.params is not None:
            base["params"] = self.params
        return base

    @staticmethod
    def _from_payload(d: dict) -> Union[DtlError, "KinesisDatastoreDef"]:
        type_field = d.get(DatastoreDef.type_field)
        if not isinstance(type_field, str):
            return DtlError("string %s is missing from the json" % (DatastoreDef.type_field))

        if type_field != KinesisDatastoreDef.type_str.value:
            return DtlError("The object %s is not an Kinesis definition" % (str(d)))

        stream_field = d.get("streamName")
        if stream_field is None:
            return DtlError("'streamName': needs to be defined")

        format_field = d.get("format")
        if format_field is None:
            return DtlError("'format': needs to be defined")
        else:
            format_field = FileFormat.file_format_from_str(format_field)
            if isinstance(format_field, DtlError):
                return format_field

        shard_attributes = d.get("shardAttributes")
        if shard_attributes is None:
            return DtlError("'shardAttributes': needs to be defined")
        else:
            shard_attributes = ShardAttributes._from_payload(shard_attributes)

        params = d.get("params")
        datastore_id = d.get(DatastoreDef.id_field)

        return KinesisDatastoreDef(stream_field, format_field, shard_attributes, params, datastore_id)


class KafkaDatastoreDef(DatastoreDef):
    type_str = CredentialType.Kafka

    """
    :param read_from_timestamp: only read records from the topic that have been published after the specified datetime, in the timezone of the kafka cluster
    """

    def __init__(
        self,
        topic: str,
        format_or_registry: Union[FileFormat, SchemaRegistry],
        params: Optional[Dict[str, str]] = None,
        datastore_id: Optional[UUID] = None,
        read_from_timestamp: Optional[datetime] = None,
    ):
        DatastoreDef.__init__(self, KafkaDatastoreDef.type_str, datastore_id)
        self.topic = topic
        self.format_or_registry = format_or_registry
        self.params = params
        self.read_from_timestamp = read_from_timestamp

    def __repr__(self):
        return (
            f"{self.__class__.__name__}(topic= {self.topic!r}, "
            f"format_or_registry= {self.format_or_registry!r}, params= {self.params}, read_from_timestamp= {self.read_from_timestamp})"
        )

    def _as_payload(self) -> dict:
        """
        Dictionary representation of the object with camelCase keys
        :return:
        """
        base = self._base_payload()
        base["topic"] = self.topic
        if isinstance(self.format_or_registry, FileFormat):
            base["format"] = self.format_or_registry.value
        else:
            base["format"] = SchemaRegistry._as_payload(self.format_or_registry)
        if self.params is not None:
            base["params"] = self.params
        if self.read_from_timestamp is not None:
            base["readFromTimestamp"] = int(self.read_from_timestamp.timestamp() * 1000)
        return base

    @staticmethod
    def _from_payload(d: dict) -> Union[DtlError, "KafkaDatastoreDef"]:
        type_field = d.get(KafkaDatastoreDef.type_field)
        if not isinstance(type_field, str):
            return DtlError("string %s is missing from the json" % (DatastoreDef.type_field))

        if type_field != KafkaDatastoreDef.type_str.value:
            return DtlError("The object %s is not a Kafka definition" % (str(d)))

        topic = d.get("topic")
        if topic is None:
            return DtlError("'topic' needs to be defined")

        format_or_registry = d.get("format")
        if format_or_registry is None:
            return DtlError("'format' needs to be defined")
        elif isinstance(format_or_registry, str):
            format_or_registry = FileFormat.file_format_from_str(format_or_registry)
            if isinstance(format_or_registry, DtlError):
                return format_or_registry
            if format_or_registry not in (
                FileFormat.Csv,
                FileFormat.Text,
                FileFormat.Json,
            ):
                return DtlError("'format' needs to be CSV, TEXT, Json")
        else:
            format_or_registry = SchemaRegistry._from_payload(format_or_registry)
            if isinstance(format_or_registry, DtlError):
                return format_or_registry

        params = d.get("params")
        resource_id = d.get(DatastoreDef.id_field)

        read_from_timestamp = d.get("readFromTimestamp")
        if read_from_timestamp is not None:
            read_from_timestamp = datetime.fromtimestamp(read_from_timestamp / 1000)
        return KafkaDatastoreDef(topic, format_or_registry, params, resource_id, read_from_timestamp)


###############################################################################
#                              API Sources
###############################################################################


class SocrataDatastoreDef(DatastoreDef):
    """
    Creates a datastore that references a Socrata dataset
    """

    type_str = CredentialType.Socrata

    def __init__(self, socrata_id: str, datastore_id: Optional[UUID] = None):
        """
        :param socrata_id: id of the dataset on socrata
        :param datastore_id: id of the datastore on the datalogue platform. This will only be set if the object was
         retrieved from the platform
        """
        DatastoreDef.__init__(self, SocrataDatastoreDef.type_str, datastore_id)
        self.socrata_id = socrata_id

    def _as_payload(self) -> dict:
        """Dictionary representation of the object with camelCase keys"""
        base = self._base_payload()
        base["socrataId"] = self.socrata_id
        return base

    @staticmethod
    def _from_payload(d: dict) -> Union[DtlError, "SocrataDatastoreDef"]:
        type_field = d.get(DatastoreDef.type_field)
        if not isinstance(type_field, str):
            return DtlError("string %s is missing from the json" % DatastoreDef.type_field)

        if type_field != SocrataDatastoreDef.type_str.value:
            return DtlError("The object %s is not a Socrata definition" % str(d))

        socrata_id = d.get("socrataId")
        if socrata_id is None:
            return DtlError("'socrata_id' needs to be defined in a Socrata definition")

        datastore_id = d.get(DatastoreDef.id_field)

        return SocrataDatastoreDef(socrata_id, datastore_id)


class HttpDatastoreDef(DatastoreDef):
    """
    Creates a datastore that references a Http dataset
    """

    type_str = CredentialType.Http

    def __init__(
        self,
        endpoint: str,
        file_format: FileFormat,
        method_type: HttpMethodType,
        headers: Optional[dict] = None,
        query_params: Optional[dict] = None,
        body: Optional[dict] = None,
        datastore_id: Optional[UUID] = None,
    ):
        """
        :param endpoint: path to attach to base url in the http credential
        :param file_format: format of the file (json, cvsv, etc.)
        :param method_type: http method selected for the request (GET, PUT, POST)
        :param headers: headers to be added to the request (Accept, X-App-Token, etc.). it's optional
        :param query_params: http request parameters. it's optional
        """
        DatastoreDef.__init__(self, HttpDatastoreDef.type_str, datastore_id)
        self.endpoint = endpoint
        self.file_format = file_format
        self.method_type = method_type
        self.headers = headers
        self.query_params = query_params
        self.body = body

    def _as_payload(self) -> dict:
        """Dictionary representation of the object with camelCase keys"""
        base = self._base_payload()
        base["endpoint"] = self.endpoint
        base["format"] = self.file_format.value
        base["methodType"] = self.method_type.value
        if self.headers is not None:
            base["headers"] = self.headers
        if self.query_params is not None:
            base["queryParams"] = self.query_params
        if self.body is not None:
            base["body"] = self.body

        return base


def _http_datastore_def_from_payload(d: dict) -> Union[DtlError, HttpDatastoreDef]:
    type_field = d.get(DatastoreDef.type_field)
    if not isinstance(type_field, str):
        return DtlError("string %s is missing from the json" % DatastoreDef.type_field)

    if type_field != HttpDatastoreDef.type_str.value:
        return DtlError("The object %s is not an HttpResource definition" % str(d))

    endpoint = d.get("endpoint")
    if endpoint is None:
        return DtlError("'endpoint' needs to be defined in an HttpResource definition")

    file_format = d.get("format")
    if file_format is None:
        return DtlError("'file_format' needs to be defined in an HttpResource definition")
    else:
        file_format = FileFormat.file_format_from_str(file_format)
        if isinstance(file_format, DtlError):
            return file_format

    method_type = d.get("methodType")
    if method_type is None:
        return DtlError("'method_type' needs to be defined in an HttpResource definition")
    else:
        method_type = HttpMethodType.method_type_from_str(method_type)
        if isinstance(method_type, DtlError):
            return method_type

    datastore_id = d.get(DatastoreDef.id_field)
    return HttpDatastoreDef(
        endpoint,
        file_format,
        method_type,
        d.get("headers"),
        d.get("queryParams"),
        d.get("body"),
        datastore_id,
    )


###############################################################################
#                              File Sources
###############################################################################


class FileDatastoreDef(DatastoreDef):
    type_str = CredentialType.FileSystem

    def __init__(
        self,
        location: str,
        file_format: FileFormat,
        params: Optional[Dict[str, str]] = None,
        datastore_id: Optional[UUID] = None,
    ):
        DatastoreDef.__init__(self, FileDatastoreDef.type_str, datastore_id)
        self.location = location
        self.file_format = file_format
        self.params = params

    def _as_payload(self) -> dict:
        """Dictionary representation of the object with camelCase keys"""
        base = self._base_payload()
        base["location"] = self.location
        base["format"] = self.file_format.value
        if self.params is not None:
            base["params"] = self.params
        return base


def _file_datastore_def_from_payload(d: dict) -> Union[DtlError, FileDatastoreDef]:
    type_field = d.get(DatastoreDef.type_field)
    if not isinstance(type_field, str):
        return DtlError("string %s is missing from the json" % DatastoreDef.type_field)

    if type_field != FileDatastoreDef.type_str.value:
        return DtlError("The object %s is not a FileSystem definition" % str(d))

    location = d.get("location")
    if location is None:
        return DtlError("'location' needs to be defined in a FileSystem definition")

    file_format = d.get("format")
    if file_format is None:
        return DtlError("'file_format' needs to be defined in a FileSystem definition")
    else:
        file_format = FileFormat.file_format_from_str(file_format)
        if isinstance(file_format, DtlError):
            return file_format

    params = d.get("params")
    datastore_id = d.get(DatastoreDef.id_field)

    return FileDatastoreDef(location, file_format, params, datastore_id)


###############################################################################
#                              The Infinite Abyss
###############################################################################


class VoidDef(DatastoreDef):
    type_str = CredentialType.Void

    def __init__(self, datastore_id: Optional[UUID] = None):
        DatastoreDef.__init__(self, VoidDef.type_str, datastore_id)

    def __repr__(self):
        return "VoidDef()"

    def _as_payload(self) -> dict:
        """Dictionary representation of the object with camelCase keys"""
        return self._base_payload()


def _void_def_from_payload(d: dict) -> Union[DtlError, VoidDef]:
    type_field = d.get(DatastoreDef.type_field)
    if not isinstance(type_field, str):
        return DtlError("string %s is missing from the json" % DatastoreDef.type_field)

    if type_field != VoidDef.type_str.value:
        return DtlError("The object %s is not a Void definition" % str(d))

    datastore_id = d.get(DatastoreDef.id_field)

    return VoidDef(datastore_id)


###############################################################################
#                              Webhook Datastore
###############################################################################


class WebhookDatastoreDef(DatastoreDef):
    type_str = CredentialType.Webhook

    def __init__(
        self,
        message_format: FileFormat,
        identifier: Optional[str] = None,
        url: Optional[str] = None,
        datastore_id: Optional[UUID] = None,
    ):
        """
        Creates a datastore that exposes an url to the Datalogue pipelining engine

        Upon creation, external applications can push data to the url via webhooks. Incoming data will be
        continually fed into any running pipeline using the datastore as a source

        :param message_format: format of the data that will be pushed by the webhook, via the FileFormat enumeration
        :param identifier: can be used to customize the generated url; will default to the UUID of the created
            datastore. Identifiers must be unique within your deployment, and spaces and special characters are not
            supported
        :param url: for backend use; upon creation, the url property will be populated with the API url
            url to which other applications can push data. User input to this param will be ignored.
        :param datastore_id: UUID of the datastore
        """
        DatastoreDef.__init__(self, WebhookDatastoreDef.type_str, datastore_id)
        self.message_format = message_format
        self.identifier = identifier
        self.url = url

    def __repr__(self):
        return (
            f"{self.__class__.__name__}(message_format= {self.message_format}, "
            f"identifier= {self.identifier!r}, url= {self.url!r}, datastore_id= {self.datastore_id!r})"
        )

    def _as_payload(self) -> dict:
        base = self._base_payload()
        # buffer is not needed in SDK, but might be introduced in upcoming versions of Webhooks
        base["buffer"] = 0
        base["fileFormat"] = self.message_format.value

        if self.identifier is not None:
            base["identifier"] = self.identifier

        if self.url is not None:
            base["webhookUrl"] = self.url

        return base

    @staticmethod
    def _webhook_def_from_payload(d: dict) -> Union[DtlError, "WebhookDatastoreDef"]:
        type_field = d.get(DatastoreDef.type_field)
        if not isinstance(type_field, str):
            return DtlError("string %s is missing from the json" % DatastoreDef.type_field)

        if type_field != WebhookDatastoreDef.type_str.value:
            return DtlError("The object %s is not a Webhook definition" % str(d))

        message_format = d.get("fileFormat")
        if not isinstance(message_format, str):
            return DtlError("'message_format' needs to be defined in a Webhook definition")
        else:
            message_format = FileFormat.file_format_from_str(message_format)

        identifier = d.get("identifier")

        url = d.get("webhookUrl")

        datastore_id = d.get(DatastoreDef.id_field)

        return WebhookDatastoreDef(message_format, identifier, url, datastore_id)


###############################################################################
#                              SFTP Datastore
###############################################################################


class SFTPDatastoreDef(DatastoreDef):
    type_str = CredentialType.SFTP

    def __init__(
        self,
        path: str,
        file_format: FileFormat,
        params: Optional[Dict[str, str]] = None,
        decompression: Optional[Decompression] = Decompression.Infer,
        resource_type: Optional[ResourceType] = None,
    ):
        """
        SFTP Datastore definition
        :param path: path within the VM to the filesystem containing data
        :param file_format: format of the file (json, csv, etc.)
        :param params: optional parameters fed to the specified file format
        :param decompression: can be used to decompress files in stream. Default behavior is to infer compression type and decompress. Supported format(s): gzip
        :param resource_type: TBD
        """
        DatastoreDef.__init__(self, SFTPDatastoreDef.type_str)
        self.path = path
        self.file_format = file_format
        self.params = params
        self.decompression = decompression
        self.resource_type = resource_type

    def __repr__(self):
        return (
            f"{self.__class__.__name__}(path= {self.path},"
            f" file_format= {self.file_format},"
            f" params= {self.params},"
            f" decompression= {self.decompression},"
            f" resource_type= {self.resource_type})"
        )

    def _as_payload(self) -> dict:
        base = self._base_payload()
        base["path"] = self.path
        base["fileFormat"] = self.file_format.value
        if self.params is not None:
            base["params"] = self.params
        base["decompression"] = self.decompression.value
        if self.resource_type is not None:
            base["resourceType"] = self.resource_type.value
        return base

    @staticmethod
    def _sftp_def_from_payload(d: dict) -> Union[DtlError, "SFTPDatastoreDef"]:
        type_field = d.get(DatastoreDef.type_field)
        if not isinstance(type_field, str):
            return DtlError("string %s is missing from the json" % DatastoreDef.type_field)

        if type_field != SFTPDatastoreDef.type_str.value:
            return DtlError("The object %s is not a SFTP datastore definition" % str(d))

        path = d.get("path")
        if not isinstance(path, str):
            return DtlError("'path' needs to be specified for SFTP datastore definition")

        file_format = d.get("fileFormat")
        if not isinstance(file_format, str):
            return DtlError("'fileFormat' needs to be specified for SFTP datastore definition")
        else:
            file_format = FileFormat.file_format_from_str(file_format)

        decompression = d.get("decompression")

        if decompression is None:
            return DtlError("'decompression' needs to be defined for SFTP datastore definition")
        else:
            decompression = Decompression.decompression_from_str(decompression)
            if isinstance(decompression, DtlError):
                return decompression

        resource_type = d.get("resourceType")
        if isinstance(resource_type, str):
            resource_type = ResourceType.resource_type_from_str(resource_type)
        if isinstance(resource_type, DtlError):
            return resource_type

        params = d.get("params")

        return SFTPDatastoreDef(path, file_format, params, decompression, resource_type)


###############################################################################
#                             ElasticSearch Datastore
###############################################################################
class ElasticSearchDef(DatastoreDef):
    type_str = CredentialType.ElasticSearch

    def __init__(
        self,
        index: str,
        index_id_path: List[str],
        type_name: str = "_doc",
        datastore_id: Optional[UUID] = None,
    ):
        DatastoreDef.__init__(self, ElasticSearchDef.type_str, datastore_id)
        self.index = index
        self.index_id_path = index_id_path
        self.type_name = type_name

    def __repr__(self):
        return (
            f"{self.__class__.__name__}(index={self.index!r},"
            f"index_id_path={self.index_id_path}"
            f"type_name={self.type_name!r})"
        )

    def _as_payload(self) -> dict:
        """Dictionary representation of the object with camelCase keys"""
        base = self._base_payload()
        base["index"] = self.index
        base["indexIdKey"] = self.index_id_path
        base["docTypeName"] = self.type_name
        return base

    @classmethod
    def _from_payload(cls, json: dict) -> Union[DtlError, "ElasticSearchDef"]:
        type_field = json.get(DatastoreDef.type_field)
        if not isinstance(type_field, str):
            return DtlError("string %s is missing from the json" % DatastoreDef.type_field)

        if type_field != cls.type_str.value:
            return DtlError("The object %s is not an ElasticSearch definition" % str(d))

        index = json.get("index")
        if index is None:
            return DtlError("'index' needs to be defined in an ElasticSearch definition")

        index_id_key = json.get("indexIdKey")
        if index_id_key is None:
            return DtlError("'indexIdKey' needs to be defined in an ElasticSearch definition")
            if not isinstance(index_id_key, list) or not all(isinstance(path, str) for path in lst):
                return DtlError("'indexIdKey' must be defined as a list of strings")

        type_name = json.get("docTypeName")
        if type_name is None:
            return DtlError("'docTypeName' needs to be defined in an ElasticSearch definition")

        datastore_id = json.get(DatastoreDef.id_field)

        return cls(
            index=index,
            index_id_path=index_id_key,
            type_name=type_name,
            datastore_id=datastore_id,
        )


###############################################################################
#                              Box Datastore
###############################################################################
class BoxDatastoreDef(DatastoreDef):
    type_str = CredentialType.Box

    def __init__(
        self,
        global_file_id: str,
        file_format: FileFormat,
        params: Optional[Dict[str, str]] = None,
        resource_type: Literal["Folder", "File"] = "File",
        datastore_id: Optional[UUID] = None,
    ):
        """
        :param global_file_id : unique identifier for box files. Commonly found in the url
        :param file_format: file format for box file
        """
        DatastoreDef.__init__(self, BoxDatastoreDef.type_str, datastore_id)
        self.global_file_id = global_file_id
        self.file_format = file_format
        self.params = params
        # TODO: add compression types
        self.resource_type = resource_type

    def __repr__(self):
        return (
            f"{self.__class__.__name__}(global_file_id={self.global_file_id!r},"
            f"file_format={self.file_format!r},params={self.params},"
            f"resource_type={self.resource_type!r})"
        )

    def _as_payload(self) -> dict:
        """
        Dictionary representation of the object with camelCase keys
        """
        base = self._base_payload()
        base["globalFileId"] = self.global_file_id
        base["format"] = self.file_format.value
        base["resourceType"] = self.resource_type
        if self.params is not None:
            base["params"] = self.params
        return base

    @classmethod
    def _from_payload(cls, d: dict) -> Union[DtlError, "BoxDatastoreDef"]:
        type_field = d.get(DatastoreDef.type_field)
        if not isinstance(type_field, str):
            return DtlError("string %s is missing from the json" % (DatastoreDef.type_field))

        if type_field != BoxDatastoreDef.type_str.value:
            return DtlError("The object %s is not a Box definition" % (str(d)))

        global_file_id = d.get("globalFileId")
        if global_file_id is None:
            return DtlError("'global_file_id' needs to be defined in a Box definition")

        file_format = d.get("format")
        if file_format is None:
            return DtlError("'file_format' needs to be defined in a Box definition")
        else:
            file_format = FileFormat.file_format_from_str(file_format)
            if isinstance(file_format, DtlError):
                return file_format

        resource_type = d.get("resourceType")
        if resource_type is None:
            resource_type = "File"

        params = d.get("params")
        datastore_id = d.get(DatastoreDef.id_field)

        return cls(
            global_file_id=global_file_id,
            file_format=file_format,
            params=params,
            resource_type=resource_type,
            datastore_id=datastore_id,
        )


_data_definitions = dict(
    [
        (S3DatastoreDef.type_str.value, _s3_datastore_def_from_payload),
        (AVCOrderItemsDatastore.type_str.value, _avc_datastore_def_from_payload),
        (GCSDatastoreDef.type_str.value, _gcs_datastore_def_from_payload),
        (AzureDatastoreDef.type_str.value, _azure_datastore_def_from_payload),
        (JdbcDatastoreDef.type_str.value, _jdbc_datastore_def_from_payload),
        (MongoDatastoreDef.type_str.value, _mongo_datastore_def_from_payload),
        (HttpDatastoreDef.type_str.value, _http_datastore_def_from_payload),
        (SocrataDatastoreDef.type_str.value, SocrataDatastoreDef._from_payload),
        (FileDatastoreDef.type_str.value, _file_datastore_def_from_payload),
        (VoidDef.type_str.value, _void_def_from_payload),
        (KinesisDatastoreDef.type_str.value, KinesisDatastoreDef._from_payload),
        (KafkaDatastoreDef.type_str.value, KafkaDatastoreDef._from_payload),
        (HadoopDatastoreDef.type_str.value, HadoopDatastoreDef._from_payload),
        (DatalogueDatastoreDef.type_str.value, DatalogueDatastoreDef._from_payload),
        (
            WebhookDatastoreDef.type_str.value,
            WebhookDatastoreDef._webhook_def_from_payload,
        ),
        (SFTPDatastoreDef.type_str.value, SFTPDatastoreDef._sftp_def_from_payload),
        (ElasticSearchDef.type_str.value, ElasticSearchDef._from_payload),
        (BoxDatastoreDef.type_str.value, BoxDatastoreDef._from_payload),
    ]
)


def _datastore_def_from_payload(json: dict) -> Union[DtlError, DatastoreDef]:
    type_field = json.get(DatastoreDef.type_field)
    if type_field is None:
        return DtlError("The datastore definition object doesn't have a '%s' property" % DatastoreDef.type_field)

    parsing_function = _data_definitions.get(type_field)
    if parsing_function is None:
        return DtlError("Looks like '%s' datastore is not handled by the SDK" % type_field)

    return parsing_function(json)


def _datastore_from_payload(json: dict) -> Union[DtlError, Datastore]:
    name = json.get("name")
    if name is None:
        return DtlError("'name' for a datastore should be defined")

    definition = json.get("definition")
    gate = json.get("gate")

    if definition is None and gate is None:
        return DtlError("Neither 'definition' nor 'gate' for a datastore is defined!'")
    elif definition is not None:
        definition = _datastore_def_from_payload(definition)
    else:
        definition = _datastore_def_from_payload(gate)

    alias = json.get("alias")
    credential_id = json.get("credentialsId")
    if credential_id is not None:
        credential_id = UUID(credential_id)
    datastore_id = json.get("id")

    schema_paths = json.get("schemaPaths")
    if schema_paths is not None:
        schema_paths = _parse_list(_parse_string_list)(schema_paths)
        if isinstance(schema_paths, DtlError):
            return schema_paths

    schema_classes = []
    classifications_json = json.get("classifications")
    if classifications_json is not None:
        classifications = _parse_list(Classification._from_payload)(classifications_json)
        if isinstance(classifications, DtlError):
            return classifications
        schema_classes = classifications

    schema_labels = None
    if schema_paths is not None:
        schema_labels = set()
        for row in schema_paths:
            for col in row:
                schema_labels.add(col)
        schema_labels = list(schema_labels)

    ads_json = json.get("ads")
    schema_nodes = None
    ads = None
    if ads_json is not None:
        schema_nodes = ads_json.get("nodes")
        if schema_nodes is not None:
            schema_nodes = _parse_list(SchemaNode._schema_node_from_payload)(schema_nodes)
            if isinstance(schema_nodes, DtlError):
                return schema_nodes
        ads = AbstractDataSchema._from_payload(ads_json, AbstractDataSchema.node_decoder)
    tags = json.get("tags")
    if tags is None:
        tags = []
    return Datastore(
        name,
        definition,
        credential_id,
        alias,
        datastore_id,
        schema_paths,
        schema_classes,
        schema_labels,
        schema_nodes,
        ads,
        tags,
    )


class ReclassificationBehavior(SerializableStringEnum):
    ReplaceAutomatic = "ReplaceAutomatic"
    ReplaceAll = "ReplaceAll"
    AddNew = "AddNew"

    @staticmethod
    def parse_error(s: str) -> DtlError:
        return DtlError(_enum_parse_error("reclassification behavior", s), DtlError(""))

    def __str__(self):
        return str(self.value)

    @staticmethod
    def reclassification_behavior_from_str(
        string: str,
    ) -> Union[DtlError, "ReclassificationBehavior"]:
        return SerializableStringEnum.from_str("ReclassificationBehavior")(string)


class RollupStrategy(ABC):
    def __init__(self):
        super().__init__()

    @abstractmethod
    def _as_payload(self) -> dict:
        """
        Dictionary representation of the object
        :return:
        """
        pass


class Threshold:
    """
    :param class_ids: List of class Ids. Can be a single class as well.
    :threshold: threshold value for class_ids. It should be between 0 and 1, [0.0, 1.0]

    Takes in a list of class ids with threshold levels to meet for each class_id. User can also enter a list of classes with one threshold to apply to all class_ids
    """

    def __init__(self, classIds: List[UUID], threshold: float):
        self.classIds = classIds
        self.threshold = threshold

    def _as_payload(self) -> dict:
        payload = {
            "classIds": list(map(lambda classId: str(classId), self.classIds)),
            "threshold": str(self.threshold),
        }
        return payload

    def __repr__(self):
        return f'(classIds= [{",".join(map(lambda t: str(t), self.classIds))}], threshold= {self.threshold})'


class ClassThreshold(RollupStrategy):
    """
    Generalizes field classification by setting thresholds for individual classes.
    Top class is selected in instances where multiple thresholds are met. No class is returned if no threshold is met

    :param threshold_pairs: Accepts the class Threshold instance() which contains class_ids and threshold values
    """

    def __init__(self, threshold_pairs: List[Threshold]):
        super().__init__()
        self.threshold_pairs = threshold_pairs

    def _as_payload(self) -> dict:
        payload = {
            "type": "Threshold",
            "thresholds": list(map(lambda threshold: threshold._as_payload(), self.threshold_pairs)),
        }
        return payload

    def __repr__(self):
        return f'Threshold(threshold_pairs= {",".join(map(lambda t: str(t), self.thresholds))})'


class WeightedAverage(RollupStrategy):
    """
    Generalizes field classification by calculating a weighted average of the individual classes
    """

    def __init__(self):
        super().__init__()

    def _as_payload(self) -> dict:
        payload = {"type": "WeightedAverage"}
        return payload

    def __repr__(self):
        return "WeightedAverage()"



from typing import Optional, Union, List
from uuid import UUID

from datalogue.errors import DtlError
from datalogue.clients._http import _HttpClient, HttpMethod
from datalogue.models.permission import Permission
from datalogue.models.regex import Regex, SearchOrder, RegexTestSample
from datalogue.models.permission import ObjectType, SharePermission
from datalogue.models.scope_level import Scope
from datalogue.dtl_utils import _parse_list


class _RegexClient:
    """
    Client to interact with the Regexes
    """

    def __init__(self, http_client: _HttpClient):
        self.http_client = http_client
        self.service_uri = "/scout"

    def create(self, regex: Regex) -> Union[DtlError, Regex]:
        """
        Creates the Regex as specified.

        :param regex: Regex definition
        :return: Regex reference with more information id, owner, regex test sample statuses
        """
        res = self.http_client.make_authed_request(
            self.service_uri + "/regexes", HttpMethod.POST, Regex._as_payload(regex)
        )

        if isinstance(res, DtlError):
            return res

        return Regex._from_payload(res)

    def update(
        self,
        id: UUID,
        name: Optional[str] = None,
        description: Optional[str] = None,
        pattern: Optional[str] = None,
    ) -> Union[Regex, DtlError]:
        """
        Update the regex as specified.

        :param id: id of regex to update
        :param name: Updated name of the regex (Optional)
        :param description: Updated description of the regex (Optional)
        :param pattern: Updated pattern of the regex (Optional)
        :return: Regex reference with more information id, owner, regex test sample statuses
        """
        req_body = {}
        if name is not None:
            req_body["name"] = name
        if description is not None:
            req_body["description"] = description
        if pattern is not None:
            req_body["pattern"] = pattern

        res = self.http_client.make_authed_request(
            self.service_uri + "/regexes/" + str(id), HttpMethod.PUT, req_body
        )

        if isinstance(res, DtlError):
            return res

        return Regex._from_payload(res)

    def share(
        self,
        regex_id: UUID,
        target_id: UUID,
        target_type: Scope,
        permission: Permission,
    ) -> Union[SharePermission, DtlError]:
        """
        Modify the sharing permission for the given regex. The change can be performed to user, group or organization with the desired permission (Share, Write or Read)
        In order to share, the user needs to have `Share` permission. Regex owners always have `Share` permission for that regex.

        :param regex_id: UUID is the id of the regex that you want to share
        :param target_id: UUID is the id of the User, Group or Organization you want to share with (depending on the target_type param)
        :param target_type: Scope (`Organization`, `Group` or `User`) with whom you want to share the regex. It can be Organization Group or User.
        :param permission: Permission (`Share`, `Write` or `Read`) the permission you want to grant
        :return: If successful returns the permission you granted specifying the `target_id` (UUID),
        the `target_type` (User, Group or Organization) and the `permission` level (Read, Write or Share) you just granted.
        (i.e. {target_id: str, target_type: str, permission: str}),
        else returns :class:`DtlError`.
        """

        url = f"/scout/regexes/{str(regex_id)}/shares?targetType={str(target_type.value)}&targetId={str(target_id)}&permission={str(permission.value)}"

        resp = self.http_client.execute_authed_request(url, HttpMethod.POST)

        if isinstance(resp, DtlError):
            return resp
        elif resp.status_code == 200:
            return SharePermission(object_type=ObjectType.Regex).from_payload(
                resp.json()
            )
        else:
            return DtlError(resp.text)

    def search(
        self,
        by_name: Optional[List[str]] = None,
        by_description: Optional[List[str]] = None,
        by_test_data: Optional[List[str]] = None,
        by_match: Optional[str] = None,
        by_owner: Optional[List[UUID]] = None,
        page: int = 1,
        results_per_page: int = 25,
    ) -> Union[List[Regex], DtlError]:
        """
        Search for a regex (full-text search), where defined (non-None) parameters/queries are the search criteria
        joint by AND operator and those not defined are ignored as search criteria.

        :param by_name: (Optional) Name of regex to be searched for (Optional)
        :param by_description: (Optional) Description of the regex to be searched for (Optional)
        :param by_test_data: (Optional) The textual test data of a regex to be searched for (Optional)
        :param by_match: (Optional) Text that can be pattern matched by the regex to be searched for (Optional)
        :param by_owner: (Optional) Regexes created by this particular owner
        :param page: The index of the page to get the regexes from. (Optional, defaults to first page index 1)
        :param results_per_page: Size per page. (Optional, defaults to 25)
        :return: List of regex references
        """
        body = {}
        query = {}
        if isinstance(by_name, list):
            query["name"] = by_name
        if isinstance(by_description, list):
            query["description"] = by_description
        if isinstance(by_test_data, list):
            query["testData.text"] = by_test_data
        if isinstance(by_owner, list):
            query["owner"] = str(by_owner)

        if isinstance(by_match, str):
            body["match"] = by_match
        if isinstance(page, int):
            body["page"] = page
        if isinstance(results_per_page, int):
            body["size"] = results_per_page
        if len(query.keys()) == 0:
            body["query"] = "*"
        else:
            body["query"] = query

        res = self.http_client.make_authed_request(
            self.service_uri + "/regexes/search", HttpMethod.POST, body=body
        )
        if isinstance(res, DtlError):
            return res

        return _parse_list(Regex._from_payload)(res)

    def test(self, id: UUID, data: List[str]) -> Union[List[RegexTestSample], DtlError]:
        """
        Test data against existing regex with given id

        :param id: id of existing regex
        :param data: List of string to verify against existing regex
        :return: data with corresponding statuses if regex with given id exists
        """
        res = self.http_client.make_authed_request(
            self.service_uri + "/regexes/" + str(id) + "/match", HttpMethod.PUT, data
        )
        if isinstance(res, DtlError):
            return res
        return _parse_list(RegexTestSample._from_payload)(res)

    def add_test_data(self, id: UUID, test_data: List[str]) -> Union[Regex, DtlError]:
        """
        Add test data to regex

        :param id: id of regex to update
        :param test_data: list of sentences which will be added as test sample
        :return: Regex reference with more information id, owner, regex test sample statuses
        """
        res = self.http_client.make_authed_request(
            self.service_uri + "/regexes/" + str(id) + "/test-data",
            HttpMethod.PUT,
            test_data,
        )
        if isinstance(res, DtlError):
            return res

        return Regex._from_payload(res)

    def remove_test_data(
        self, id: UUID, test_data: List[str]
    ) -> Union[Regex, DtlError]:
        """
        Remove test data from regex

        :param id: id of regex to update
        :param test_data: list of sentences which will be removed from test sample
        :return: Regex reference with more information id, owner, regex test sample statuses
        """
        res = self.http_client.make_authed_request(
            self.service_uri + "/regexes/" + str(id) + "/test-data",
            HttpMethod.DELETE,
            test_data,
        )
        if isinstance(res, DtlError):
            return res

        return Regex._from_payload(res)

    def get(self, id: Union[str, UUID]) -> Union[DtlError, Regex]:
        """
        Retrieve a regex by its ID.
        :param regex_id: the id of the regex to be retrieved locally
        :return: the regex if successful, or a DtlError if failed
        """
        res = self.http_client.make_authed_request(
            self.service_uri + "/regexes/" + str(id), HttpMethod.GET
        )
        if isinstance(res, DtlError):
            return res

        return Regex._from_payload(res)

    def list(
        self, page: int = 1, item_per_page: int = 25
    ) -> Union[DtlError, List[Regex]]:
        """
        List all regexes.

        :param page: page to be retrieved
        :param item_per_page: number of items to be put in a page
        :return: Returns a List of all the available regex if successful, or DtlError if failed
        """
        req_params = {}
        if isinstance(page, int):
            req_params["page"] = page
        if isinstance(item_per_page, int):
            req_params["item_per_page"] = item_per_page
        res = self.http_client.make_authed_request(
            self.service_uri + "/regexes", HttpMethod.GET, params=req_params
        )
        if isinstance(res, DtlError):
            return res

        return _parse_list(Regex._from_payload)(res)

    def delete(self, regex_id: Union[str, UUID]) -> Union[DtlError, bool]:
        """
        Delete a regex.

        :param regex_id: id of the regex to be deleted
        :return: True if successful, or DtlError if failed
        """

        resp = self.http_client.execute_authed_request(
            self.service_uri + "/regexes/" + str(regex_id), HttpMethod.DELETE
        )

        if isinstance(resp, DtlError):
            return resp
        elif resp.status_code == 200:
            return True
        else:
            return DtlError(resp.text)



from typing import List, Union, Optional, Dict

from datalogue.models.permission import (
    Permission,
    Scope,
    SharePermission,
    ObjectType,
    UnsharePermission,
)
from datalogue.clients._http import _HttpClient, HttpMethod
from datalogue.models.pipeline import *
from datalogue.models.pipeline_behavior import *
from datalogue.models.job import Job
from datalogue.errors import DtlError, _invalid_pagination_params
from uuid import UUID, uuid4
from datetime import datetime, timedelta
from dateutil.tz import UTC
from datalogue.dtl_utils import _parse_list, ResponseStream
from pyarrow import csv, Table

from datalogue.models.pipeline_template import PipelineTemplate
from datalogue.models.datastore import VoidDef
from datalogue.clients.datastore import DatastoreClient


class PipelineClient:
    def __init__(self, http_client: _HttpClient):
        self.http_client = http_client
        self.service_uri = "/scout"
        self.template = _TemplateClient(http_client)

    def create(self, name: str, pipeline_def: PipelineDef, is_resilient: bool = False) -> Union[DtlError, Pipeline]:
        """
        Creates a Pipeline with unique id

        :param is_resilient: True if the job of this pipeline will be retried
        :param name: String, a name for your pipeline
        :param pipeline_def: PipelineDef is the definition of the Pipeline
        :return: created Pipeline if successful, DtlError if failed
        """
        payload = {
            "name": str(name),
            "definition": pipeline_def._as_payload(),
            "resilient": is_resilient,
        }

        res = self.http_client.make_authed_request(self.service_uri + "/streams", HttpMethod.POST, payload)

        if isinstance(res, DtlError):
            return DtlError("Encountered an error while creating the pipeline", res)

        return Pipeline._from_payload(res)

    def get(self, pipeline_id: Union[str, UUID]) -> Union[DtlError, Pipeline]:
        """
        Retrieves a Pipeline

        :param pipeline_id: id of Pipeline to be retrieved
        :return: Pipeline if successful, DtlError if failed
        """
        uri = f"{self.service_uri}/streams/{str(pipeline_id)}"
        res = self.http_client.execute_authed_request(uri, HttpMethod.GET)

        if isinstance(res, DtlError):
            return res
        elif res.status_code == 200:
            return Pipeline._from_payload(res.json())
        if res.status_code == 400:
            payload = res.json()
            reason = payload.get("error", payload.get("message"))
            return DtlError(
                f"Encountered an error while getting the pipeline! pipeline_id: {pipeline_id}, reason: {reason}"
            )
        else:
            return self.http_client._handle_status_code(uri, res)

    def update(
        self,
        pipeline_id: Union[str, UUID],
        name: Optional[str] = None,
        pipeline_def: Optional[PipelineDef] = None,
        is_resilient: Optional[bool] = None,
    ) -> Union[DtlError, Pipeline]:
        """
        Updates the Pipeline's configuration

        :param pipeline_id: id of Pipeline to update
        :param name: updated name of Pipeline (Optional)
        :param pipeline_def: updated definition of Pipeline (Optional)
        :return: updated Pipeline if successful, DtlError if failed
        """
        payload = {}

        if name is not None:
            payload["name"] = str(name)

        if pipeline_def is not None:
            payload["definition"] = pipeline_def._as_payload()

        if is_resilient is not None:
            payload["resilient"] = is_resilient

        res = self.http_client.make_authed_request(
            f"{self.service_uri}/streams/{str(pipeline_id)}", HttpMethod.POST, payload
        )

        if isinstance(res, DtlError):
            return DtlError(
                f"Encountered an error while updating the pipeline for {pipeline_id}",
                res,
            )

        return Pipeline._from_payload(res)

    def delete(self, pipeline_id: Union[str, UUID]) -> Union[DtlError, bool]:
        """
        Deletes all instances of Pipeline

        :param pipeline_id: id of Pipeline to be deleted
        :return: True if successful, DtlError if failed
        """
        uri = f"{self.service_uri}/streams/{str(pipeline_id)}"
        res = self.http_client.execute_authed_request(uri, HttpMethod.DELETE)
        if isinstance(res, DtlError):
            return res
        elif res.status_code == 200:
            return True
        elif res.status_code == 400:
            payload = res.json()
            return DtlError(
                f'Encountered an error while deleting the pipeline! pipeline_id: {pipeline_id}, reason: {payload["error"]}'
            )
        else:
            return self.http_client._handle_status_code(uri, res)

    def list(self, page: int = 1, item_per_page: int = 25) -> Union[DtlError, List[Pipeline]]:
        """
        Retrieves a paginated array of Pipelines

        :param page: page of Pipeline list to return
        :param item_per_page: number of Pipelines to return per page
        :return: Array of Pipelines if successful, DtlError if failed
        """
        params = {"page": str(page), "size": str(item_per_page)}

        res = self.http_client.make_authed_request(f"{self.service_uri}/streams", HttpMethod.GET, params=params)

        if isinstance(res, DtlError):
            return DtlError(f"Encountered an error while listing pipelines", res)
        entities = res["entities"]
        return _parse_list(Pipeline._from_payload)(entities)

    def orchestrate(self, pipeline_id: Union[str, UUID], wait_on: List[Union[str, UUID]] = []) -> Union[DtlError, Job]:
        """
        Orchestrate the running of a pipeline to trigger after the success of one or more separate jobs

        :param :pipeline_id: id of the pipeline to be run
        :param :wait_on: id(s) of jobs which need to successfully finish before the specified pipeline will run.
        Default value is empty string which means if user does not define, there will be no dependencies
        :return: Returns Job object or error
        """
        params = {"wait_on": wait_on}

        res = self.http_client.make_authed_request(
            f"{self.service_uri}/streams/{pipeline_id}/orchestrate",
            HttpMethod.POST,
            params=params,
        )

        if isinstance(res, DtlError):
            return res

        jobs = _parse_list(Job._from_payload_v1)(res)
        if isinstance(jobs, DtlError):
            return jobs
        elif len(jobs) == 0:
            return DtlError("Encountered an issue! There should be at least one job!")
        else:
            # API returns main jobs and additional jobs per target to be analyzed once this stream is completed.
            # In SDK, we're returning only the main job to the user
            return jobs[0]

    def schedule(
        self,
        pipeline_id: Union[str, UUID],
        run_date: datetime,
        frequency: Optional[timedelta] = None,
        abort_if_overlong: bool = False,
    ) -> Union[DtlError, Job]:
        """
        Schedules the pipeline to run at the given date

        :param pipeline_id: UUID id of the pipeline to schedule
        :param date: datetime for scheduling
        :param frequency: interval value (e.g float) with period (e.g hours) after which a job will be repeated
        :param abort_if_overlong: if True, avoids running a job cycle if previous cycle is running. (False by default)
        :return: Returns Job object or error
        """
        uri = f"{self.service_uri}/streams/{pipeline_id}/schedule"
        payload = {
            "runDate": run_date.astimezone(UTC).strftime("%Y-%m-%dT%H:%M:%SZ"),
            "abortIfOverlong": abort_if_overlong,
        }
        if frequency is not None:
            payload.update({"intervalInSec": frequency.total_seconds()})
        res = self.http_client.execute_authed_request(uri, HttpMethod.POST, payload)

        if isinstance(res, DtlError):
            return res
        elif res.status_code == 200:
            jobs = _parse_list(Job._from_payload_v1)(res.json())
            if isinstance(jobs, DtlError):
                return jobs
            elif len(jobs) == 0:
                return DtlError("Encountered an issue! There should be at least one job!")
            else:
                return jobs[0]
        elif res.status_code == 400:
            payload = res.json()
            return DtlError(
                f'Encountered an error while scheduling the pipeline! pipeline_id: {pipeline_id}, reason: {payload["error"]}'
            )
        else:
            return self.http_client._handle_status_code(uri, res)

    def run(self, pipeline_id: Union[str, UUID]) -> Union[DtlError, Job]:
        """
        Runs the pipeline right now

        :param pipeline_id: id of the pipeline to run
        :return: Returns Job object or error
        """

        uri = f"{self.service_uri}/streams/{pipeline_id}/run"

        res = self.http_client.execute_authed_request(uri, HttpMethod.POST)

        if isinstance(res, DtlError):
            return res
        elif res.status_code == 200:
            jobs = _parse_list(Job._from_payload_v1)(res.json())
            if isinstance(jobs, DtlError):
                return jobs
            elif len(jobs) == 0:
                return DtlError("Encountered an issue! There should be at least one job!")
            else:
                return jobs[0]
        elif res.status_code == 400:
            payload = res.json()
            return DtlError(
                f'Encountered an error while scheduling the pipeline! pipeline_id: {pipeline_id}, reason: {payload["error"]}'
            )
        else:
            return self.http_client._handle_status_code(uri, res)

    # #TODO we need to use sth else than target_id to differentiate definitions
    def preview(
        self,
        source_id: Union[str, UUID],
        transformations: List[Transformation] = [],
        size: int = 10,
    ) -> Union[DtlError, Table]:
        """
        Returns the preview of the source data against the transformations applied.
        :param source_id: id for the source dataset
        :param transformations: a list of transformations
        :param size: the size of the output table, in records
        """
        # create a dummy target_ds object
        source_object = DatastoreClient(self.http_client)
        source_ds = source_object.get(source_id)
        target_ds = Datastore("Dummy target datasource object", VoidDef(), datastore_id=uuid4())
        # build pipeline definition for the source data  to be previewed
        pipeline_def = PipelineDef.simple_builder(
            source_datastore=source_ds,
            transformation_list=transformations,
            destination_datastore=target_ds,
            prepend_structure=False,
        )
        params = {
            "size": size,
            "file-format": "Csv",
            "sample-store-id": str(target_ds.id),
        }
        res = self.http_client.make_authed_request(
            f"{self.service_uri}/streams/samples",
            HttpMethod.POST,
            body=pipeline_def._as_payload(),
            params=params,
            stream=True,
        )
        if isinstance(res, DtlError):
            return res
        response_stream = ResponseStream(res.iter_content(1024))
        return csv.read_csv(response_stream)

    def templatize(self, pipeline_id: Union[str, UUID]) -> Union[DtlError, PipelineTemplate]:
        """
        Generate a PipelineTemplate with logic matching that of the specified pipeline,
        and with an auto-generated name.
        Once generated, the PipelineTemplate can be tweaked with any automatic or user
        defined-variables as desired, or saved directly to the backend via
        dtl.pipeline.template.create()

        :param pipeline_id: id of the pipeline for which a template is to be generated
        :return: PipelineTemplate object with auto-generated name if successful,
        DtlError otherwise
        """
        uri = f"{self.service_uri}/streams/{pipeline_id}/templatize"

        res = self.http_client.make_authed_request(uri, HttpMethod.POST)

        if isinstance(res, DtlError):
            return res

        else:
            pipeline_name = res.get("streamName")
            template_name = f"Pipeline Template from: {pipeline_name} ({pipeline_id})"
            template_yaml = res.get("templateYaml")
            template_obj = PipelineTemplate(template_name, template_yaml)
            return template_obj

    def update_schedule(
        self,
        pipeline_id: Union[str, UUID],
        run_date: Optional[datetime] = None,
        frequency: Optional[timedelta] = None,
        abort_if_overlong: Optional[bool] = None,
    ) -> Union[DtlError, Job]:

        """
        Updates the schedule for a pipeline.
        The next scheduled job will reflect the new schedule for the given pipeline ID. Pipelines running on a recurring
        schedule will retain their parent job ID and the update will affect all future runs of the pipeline.

        :param pipeline_id: UUID id of the pipeline to schedule
        :param run_date: optional, datetime for scheduling
        :param frequency: optional, interval value (e.g float) with period (e.g hours) after which a job will be repeated
        :param abort_if_overlong: optional, if True, avoids running a job cycle if previous cycle is running. (False by default)
        :return: Returns Job object or error
        :return: First scheduled job of the pipeline’s running if
        successful, or DtlError if failed
        """

        url = f"{self.service_uri}/streams/{str(pipeline_id)}/schedule"

        payload = {}
        if run_date is not None:
            payload.update({"runDate": run_date.astimezone(UTC).strftime("%Y-%m-%dT%H:%M:%SZ")})
        if frequency is not None:
            payload.update({"intervalInSec": frequency.total_seconds()})
        if abort_if_overlong is not None:
            payload.update({"abortIfOverlong": abort_if_overlong})
        if not payload:
            return DtlError("Please specify one of these params: 'run_date', 'frequency' or 'abort_if_overlong'")

        rsp = self.http_client.make_authed_request(url, HttpMethod.PATCH, payload)

        if isinstance(rsp, DtlError):
            return DtlError(f"Encountered an error while updating pipeline schedule: {rsp.message}")

        updated_job = Job._from_payload(rsp)
        return updated_job


class _TemplateClient:
    """
    Client to interact with Pipeline Templates
    """

    def __init__(self, http_client: _HttpClient):
        self.http_client = http_client
        self.service_uri = "/scout"

    def create(self, pipeline_template: PipelineTemplate, validate: bool = True) -> Union[DtlError, PipelineTemplate]:
        """
        Create and validate the structure of a pipeline template

        :param pipeline_template: a local pipeline template, to be created
        and saved on the platform
        :param validate: can be set to False to opt-out of template
        validation upon creation. Otherwise, the argument of
        dummy_input_map in the PipelineTemplate object will be used to
        complete the template
        :return: Returns created PipelineTemplate if successful, or DtlError if failed
        """

        if validate is False:
            pipeline_template.dummy_input_map = None

        if validate is True and pipeline_template.dummy_input_map is None:
            return DtlError("dummy_input_map must have key -> value if 'validate' argument is True")

        res = self.http_client.make_authed_request(
            f"{self.service_uri}/templates",
            HttpMethod.POST,
            pipeline_template._as_payload(),
        )

        if isinstance(res, DtlError):
            return res

        return PipelineTemplate._from_payload(res)

    def update(
        self,
        id: Union[str, UUID],
        name: Optional[str] = None,
        template: Optional[str] = None,
        dummy_input_map: Optional[Dict[str, str]] = None,
        validate: bool = True,
    ) -> Union[DtlError, PipelineTemplate]:
        """
        Update a template's name and/or template by its ID.
        :param id: id of the template to be updated
        :param name : optional, new name of the template
        :param template: optional, new template to update the existing template
        :param dummy_input_map:  an example of data that will be used to
        validate this template upon creation, unless disabled upon the
        creation step
        :param validate: can be set to False to opt-out of template
        validation upon update. Otherwise, the argument of
        dummy_input_map will be used to
        complete the template
        :return: Returns an updated template if successful, ors DtlError if failed
        """
        req_body = {}
        if name is not None:
            req_body["name"] = name
        if template is not None:
            req_body["template"] = template
        if dummy_input_map is not None:
            req_body["dummy_input_map"] = dummy_input_map

        if validate is False:
            req_body["dummy_input_map"] = None
        if validate is True and dummy_input_map is None:
            return DtlError("dummy_input_map must have key -> value if 'validate' argument is True")
        if dummy_input_map is not None and template is None:
            return DtlError("template must be supplied for dummy_input_map validation")

        res = self.http_client.make_authed_request(f"{self.service_uri}/templates/{str(id)}", HttpMethod.PUT, req_body)

        if isinstance(res, DtlError):
            return res

        return PipelineTemplate._from_payload(res)

    def list(self, page: int = 1, item_per_page: int = 25) -> Union[DtlError, List[PipelineTemplate]]:
        """
        List all templates.

        :param page: page to be retrieved
        :param item_per_page: number of templates to be put in a page
        :return: Returns a List of all the available templates if successful, or DtlError if failed
        """
        req_params = {}
        if isinstance(page, int) and page > 0:
            req_params["page"] = page
        else:
            return _invalid_pagination_params("page")
        if isinstance(item_per_page, int) and item_per_page > 0:
            req_params["size"] = item_per_page
        else:
            return _invalid_pagination_params("item_per_page")
        res = self.http_client.make_authed_request(f"{self.service_uri}/templates", HttpMethod.GET, params=req_params)
        if isinstance(res, DtlError):
            return res

        return _parse_list(PipelineTemplate._from_payload)(res)

    def get(self, id: Union[str, UUID]) -> Union[DtlError, PipelineTemplate]:
        """
        Retrieve a template by its ID.
        :param id: the id of the template to be retrieved locally
        :return: Returns a template if successful, or a DtlError if failed
        """
        res = self.http_client.make_authed_request(f"{self.service_uri}/templates/{str(id)}", HttpMethod.GET)
        if isinstance(res, DtlError):
            return res

        return PipelineTemplate._from_payload(res)

    def delete(self, id: Union[str, UUID]) -> Union[DtlError, bool]:
        """
        Delete a template by its ID.
        :param id: id of the template to be deleted
        :return: True if successful, or DtlError if failed
        """
        res = self.http_client.execute_authed_request(f"{self.service_uri}/templates/{str(id)}", HttpMethod.DELETE)

        if isinstance(res, DtlError):
            return res
        elif res.status_code == 204:
            return True
        else:
            return DtlError(res.text)

    def share(
        self, id: UUID, target_id: UUID, target_type: Scope, permission: Permission
    ) -> Union[SharePermission, DtlError]:
        """
        Extends Permissions for Pipeline to another User or Group

        :param id: id of Pipeline to be shared
        :param target_type: User or Group, as an enumeration of Scope
        :param target_id: id of the Permission-receiving User or Group
        :param permission: level of Permissions to share (Read, Write, or Share), as an enumeration of Permission
        :return: SharePermission if successful, DtlError if failed
        """

        url = f"{self.service_uri}/templates/{str(id)}/shares?targetType={str(target_type.value)}&targetId={str(target_id)}&permission={str(permission.value)}"
        rsp = self.http_client.execute_authed_request(url, HttpMethod.POST)

        if isinstance(rsp, DtlError):
            return rsp
        elif rsp.status_code == 200:
            return SharePermission(object_type=ObjectType.PipelineTemplate).from_payload(rsp.json())
        else:
            return DtlError(rsp.text)

    def unshare(
        self, id: UUID, target_id: UUID, target_type: Scope, permission: Permission
    ) -> Union[UnsharePermission, DtlError]:
        """
        Withdraws Permissions for Pipeline with another User or Group

        :param id: id of Pipeline to be unshared
        :param target_type: User or Group, as an enumeration of Scope
        :param target_id: id of the Permission-withdrawn User or Group
        :param permission: level of Permissions to unshare (Read, Write, or Share), as an enumeration of Permission
        :return: UnsharePermission if successful, DtlError if failed
        """

        url = f"{self.service_uri}/templates/{str(id)}/shares?targetType={str(target_type.value)}&targetId={str(target_id)}&permission={str(permission.value)}"
        rsp = self.http_client.execute_authed_request(url, HttpMethod.DELETE)

        if isinstance(rsp, DtlError):
            return rsp
        elif rsp.status_code == 200:
            return UnsharePermission(object_type=ObjectType.PipelineTemplate).from_payload(rsp.json())
        else:
            return DtlError(rsp.text)

    def validate(self, id: UUID, dummy_input_map: Dict[str, str]) -> Union[bool, DtlError]:
        """
        Validate if an existing pipeline template compiles, with a supplied map of values.
        :param id: id of the template that is to be validated
        :param dummy_input_map: an example of data that will be used to
        validate the existing template
        :return: Returns True if template is valid, or DtlError if the validation was unsuccessful
        """
        url = f"{self.service_uri}/templates/{str(id)}/validate"
        rsp = self.http_client.execute_authed_request(url, HttpMethod.POST, dummy_input_map)

        if isinstance(rsp, DtlError):
            return rsp
        elif rsp.status_code == 200:
            return True
        else:
            return DtlError(rsp.text)

    def generate_pipeline(
        self,
        pipeline_template_id: Union[str, UUID],
        pipeline_name: str,
        input_map: Dict[str, str],
    ) -> Union[Pipeline, DtlError]:
        """
        Generate a pipeline by providing the missing inputs and their
        mapping to an existing pipeline template

        The correct number and mapping of inputs must be provided to the
        template—i.e. for a template with a {{local_path}}, a mapping of
        {"local_path" : "/data/source.csv"} would be appropriate

        :param pipeline_template_id: id of the pipeline template
        :param pipeline_name: name of the pipeline to be generated
        :param input_map: a mapping of inputs to be slotted into the template,
        following moustache distinction <https://mustache.github.io/mustache.5.html>
        :return: the newly created Pipeline if successful, or DtlError if
        failed
        """
        url = f"{self.service_uri}/templates/{str(pipeline_template_id)}/pipeline"
        payload = {"name": pipeline_name, "inputMap": input_map}
        rsp = self.http_client.make_authed_request(url, HttpMethod.POST, payload)

        if isinstance(rsp, DtlError):
            return DtlError(f"Encountered an error while generating pipeline from template: {rsp.message}")

        return Pipeline._from_payload(rsp)

    def run(
        self,
        pipeline_template_id: Union[str, UUID],
        pipeline_name: str,
        input_map: Dict[str, str],
    ) -> Union[DtlError, Job]:
        """
        Run a pipeline completed by providing the missing inputs and their
        mapping to an existing pipeline template

        The correct number and mapping of inputs must be provided to the
        template—i.e. for a template with a {{local_path}}, a mapping of
        {"local_path" : "/data/source.csv"} would be appropriate

        :param pipeline_template_id: id of the pipeline template
        :param pipeline_name: name of the pipeline to be run
        :param input_map: a mapping of inputs to be slotted into the template,
        following moustache distinction <https://mustache.github.io/mustache.5.html>
        :return: a Job generated by running the completed pipeline if
        successful, or DtlError if failed
        """

        url = f"{self.service_uri}/templates/{str(pipeline_template_id)}/run"
        payload = {"name": pipeline_name, "inputMap": input_map}

        rsp = self.http_client.make_authed_request(url, HttpMethod.POST, payload)

        if isinstance(rsp, DtlError):
            return DtlError(f"Encountered an error while running pipeline from template: {rsp.message}")

        jobs = _parse_list(Job._from_payload_v1)(rsp)
        if len(jobs) == 0:
            return DtlError("Encountered an issue! There should be at least one job!")
        else:
            return jobs[0]

    def schedule(
        self,
        pipeline_template_id: Union[str, UUID],
        pipeline_name: str,
        input_map: Dict[str, str],
        run_date: datetime,
        frequency: Optional[timedelta] = None,
        abort_if_overlong: bool = False,
    ) -> Union[DtlError, Job]:

        """
        Schedule a template to generate jobs, optionally following a
        recurrent frequency

        :param pipeline_template_id: id of the pipeline template
        :param pipeline_name: name of the pipeline to be run
        :param input_map: a mapping of inputs to be slotted into the
        template, following moustache distinction
        :param run_date: localized date at which to schedule the job
        :param frequency: frequency of subsequent re-runnings, via the
        timedelta class
        :param abort_if_overlong: defaults to run a subsequent running if
        a current job is still running
        :return: First scheduled job of the pipeline’s running if
        successful, or DtlError if failed
        """

        url = f"{self.service_uri}/templates/{str(pipeline_template_id)}/schedule"

        payload = {
            "name": pipeline_name,
            "runDate": run_date.astimezone(UTC).strftime("%Y-%m-%dT%H:%M:%SZ"),
            "inputMap": input_map,
            "abortIfOverlong": abort_if_overlong,
        }

        if frequency is not None:
            payload.update({"intervalInSec": frequency.total_seconds()})

        rsp = self.http_client.make_authed_request(url, HttpMethod.POST, payload)

        if isinstance(rsp, DtlError):
            return DtlError(f"Encountered an error while running pipeline from template: {rsp.message}")

        jobs = _parse_list(Job._from_payload_v1)(rsp)
        if len(jobs) == 0:
            return DtlError("Encountered an issue! There should be at least one job!")
        else:
            return jobs[0]



from uuid import UUID

from datalogue.errors import DtlError, _enum_parse_error, _invalid_parameter_error
from datalogue.dtl_utils import _parse_string_list, SerializableStringEnum
import abc
from typing import Union, List, Optional


class Transformation(abc.ABC):
    type_field = "type"

    def __init__(self, transformation_type: str):
        self.type = transformation_type
        super().__init__()

    def _base_payload(self) -> dict:
        return dict([(Transformation.type_field, self.type)])

    @abc.abstractmethod
    def _as_payload(self) -> dict:
        """
        Represents the transformation in its dictionary construction

        :return:
        """


class RegexTransformation(Transformation):
    def __init__(
        self,
        regex: Optional[str],
        regex_id: Optional[UUID],
        regex_transformation_type: str,
    ):
        Transformation.__init__(self, regex_transformation_type)
        if regex is None and regex_id is None:
            raise _invalid_parameter_error(
                "neither regex nor regex_id is/are not defined!",
                regex_transformation_type,
            )
        if regex is not None and regex_id is not None:
            raise _invalid_parameter_error(
                "both regex and regex_id are defined! Please provide only either.",
                regex_transformation_type,
            )


class DataType(SerializableStringEnum):
    """
    Data Types that can be specified in the pipeline
    """

    String = "String"
    Boolean = "Boolean"
    Integer = "Integer"
    DateTime = "DateTime"
    Double = "Double"

    @staticmethod
    def parse_error(s: str) -> str:
        return _enum_parse_error("data type", s)

    @staticmethod
    def from_str(string: str) -> Union[DtlError, "DataType"]:
        return SerializableStringEnum.from_str(DataType)(string)


def _array_from_dict(
    json: dict, transformation_type: str, array_field_key: str
) -> Union[DtlError, List[str]]:
    """
    Generic method to extract an array from a json transformation dictionary

    :param json: dictionary built from json
    :param transformation_type: type of the transformation
    :param array_field_key: key at which the array is
    :return:
    """
    if json.get(Transformation.type_field) != transformation_type:
        return DtlError("Dictionary input is not of type %s" % transformation_type)

    array_field = json.get(array_field_key)
    if array_field is None:
        return DtlError(
            "'%s' is missing from the json transformation" % array_field_key
        )

    array = _parse_string_list(array_field)

    if isinstance(array, DtlError):
        return array

    return array


class ByClass:
    def __init__(self, class_id: Union[str, UUID]):
        self.class_id = class_id
        """
        Targets data classified as the specified class
        :param class_id: id of the class with which to specify as a transformation target
        """

    def __repr__(self):
        return "ByClass(" f"class_id= {self.class_id!r})"


class ByFieldName:
    def __init__(self, field_name: List[str]):
        self.field_name = field_name
        """
        Targets data residing under the specified field name
    
        :param field_name: name of the field to specify as a transformation target
        """

    def __repr__(self):
        return "ByFieldName(" f"field_name= {self.field_name!r})"



import abc
from typing import Union

from datalogue.errors import DtlError
from datalogue.models.datastore import EncryptionType


class Encryption(abc.ABC):
    type_field = "type"

    def __init__(self, definition_type: EncryptionType):
        self.type = definition_type

    def _base_payload(self) -> dict:
        return dict([(Encryption.type_field, self.type.value)])

    def __eq__(self, other: "Encryption"):
        if isinstance(self, other.__class__):
            return self._as_payload() == other._as_payload()
        return False

    @abc.abstractmethod
    def _as_payload(self) -> dict:
        """
        Dictionary representation of the object
        :return:
        """


class AES256(Encryption):
    def __init__(self):
        Encryption.__init__(self, EncryptionType.AES256)

    def _as_payload(self):
        return self._base_payload()


class KeyId(Encryption):
    def __init__(self, key_id: str):
        Encryption.__init__(self, EncryptionType.KeyId)
        self.keyId = key_id

    def _as_payload(self) -> dict:
        base = self._base_payload()
        base["keyId"] = self.keyId
        return base


class CustomerKey(Encryption):
    def __init__(self, key: str):
        Encryption.__init__(self, EncryptionType.CustomerKey)
        self.key = key

    def _as_payload(self) -> dict:
        base = self._base_payload()
        base["key"] = self.key
        return base
